import argparse
import glob
import json
import os
import sys
import subprocess
import time
import kdtree
import numpy as np
import pandas as pd
import seaborn as sns
import threadpoolctl
import yaml

from pprint import pprint
from matplotlib import pyplot as plt
from memory_profiler import memory_usage
from sklearn import set_config
from sklearn.neighbors import KDTree
from sklearn.utils._openmp_helpers import _openmp_effective_n_threads


# Be gentle with eyes
plt.rcParams["figure.dpi"] = 200


def benchmark(config, results_folder, bench_name):
    datasets = config["datasets"]
    estimators = config["estimators"]
    leaf_sizes = config["leaf_sizes"]
    n_neighbors = config.get("n_neighbors", [])

    n_trials = config.get("n_trials", 3)
    return_distance = config.get("return_distance", False)
    one_GiB = 1e9
    benchmarks = pd.DataFrame()

    n_threads = int(_openmp_effective_n_threads())

    env_specs_file = f"{results_folder}/{bench_name}.json"

    # TODO: This is ugly, but I haven't found something better.
    commit = (
        str(subprocess.check_output(["git", "rev-parse", "--short", "HEAD"]))
        .replace("b'", "")
        .replace("\\n'", "")
    )

    env_specs = dict(
        threadpool_info=threadpoolctl.threadpool_info(),
        commit=commit,
        config=config,
        n_threads=n_threads,
    )

    set_config(assume_finite=True)

    with open(env_specs_file, "w") as outfile:
        json.dump(env_specs, outfile)

    for dataset in datasets:
        for leaf_size in leaf_sizes:
            for trial in range(n_trials):
                dataset = {k: int(float(v)) for k, v in dataset.items()}
                ns_train, ns_test, n_features = dataset.values()
                X_train = np.random.rand(ns_train, n_features)
                X_test = np.random.rand(ns_test, n_features)
                bytes_processed_data_init = X_train.nbytes
                bytes_processed_data_query = X_test.nbytes

                t0_ = time.perf_counter()
                sk_tree = KDTree(X_train, leaf_size=leaf_size)
                t1_ = time.perf_counter()
                time_elapsed = round(t1_ - t0_, 5)

                row = dict(
                    trial=trial,
                    func="init",
                    implementation="sklearn",
                    n_threads=n_threads,
                    leaf_size=leaf_size,
                    n_samples_train=ns_train,
                    n_samples_test=ns_test,
                    n_features=n_features,
                    n_neighbors=np.nan,
                    time_elapsed=time_elapsed,
                    throughput=bytes_processed_data_init / time_elapsed / one_GiB,
                )

                benchmarks = benchmarks.append(row, ignore_index=True)
                pprint(row)
                print("---")

                t0_ = time.perf_counter()
                tree = kdtree.KDTree(X_train, leaf_size=leaf_size)
                t1_ = time.perf_counter()
                time_elapsed = round(t1_ - t0_, 5)

                row = dict(
                    trial=trial,
                    func="init",
                    implementation="kdtree",
                    n_threads=n_threads,
                    leaf_size=leaf_size,
                    n_samples_train=ns_train,
                    n_samples_test=ns_test,
                    n_features=n_features,
                    n_neighbors=np.nan,
                    time_elapsed=time_elapsed,
                    throughput=bytes_processed_data_init / time_elapsed / one_GiB,
                )

                benchmarks = benchmarks.append(row, ignore_index=True)
                pprint(row)
                print("---")

                benchmarks.to_csv(
                    f"{results_folder}/{bench_name}.csv",
                    mode="w+",
                    index=False,
                )

                for k in n_neighbors:

                    t0_ = time.perf_counter()
                    sk_tree.query(X_test, k=k, return_distance=False)
                    t1_ = time.perf_counter()
                    time_elapsed = round(t1_ - t0_, 5)

                    row = dict(
                        trial=trial,
                        func="query",
                        implementation="sklearn",
                        leaf_size=leaf_size,
                        n_samples_train=ns_train,
                        n_samples_test=ns_test,
                        n_features=n_features,
                        n_neighbors=k,
                        time_elapsed=time_elapsed,
                        throughput=bytes_processed_data_query / time_elapsed / one_GiB,
                    )

                    benchmarks = benchmarks.append(row, ignore_index=True)
                    pprint(row)
                    print("---")

                    closests = np.zeros((ns_test, k), dtype=np.int32)
                    t0_ = time.perf_counter()
                    tree.query(X_test, closests)
                    t1_ = time.perf_counter()
                    time_elapsed = round(t1_ - t0_, 5)

                    row = dict(
                        trial=trial,
                        func="query",
                        implementation="kdtree",
                        leaf_size=leaf_size,
                        n_samples_train=ns_train,
                        n_samples_test=ns_test,
                        n_features=n_features,
                        n_neighbors=k,
                        time_elapsed=time_elapsed,
                        throughput=bytes_processed_data_query / time_elapsed / one_GiB,
                    )

                    benchmarks = benchmarks.append(row, ignore_index=True)
                    pprint(row)
                    print("---")

                    benchmarks.to_csv(
                        f"{results_folder}/{bench_name}.csv",
                        mode="w+",
                        index=False,
                    )



    # Overriding again now that all the dyn. lib. have been loaded
    env_specs["threadpool_info"] = threadpoolctl.threadpool_info()

    with open(env_specs_file, "w") as outfile:
        json.dump(env_specs, outfile)


if __name__ == "__main__":
    parser = argparse.ArgumentParser("benchmark")

    parser.add_argument("bench_name")

    args = parser.parse_args()

    bench_name = args.bench_name
    with open("benchmarks/config.yml", "r") as f:
        config = yaml.full_load(f)

    results_folder = f"benchmarks/results/{bench_name}"
    os.makedirs(results_folder, exist_ok=True)

    print(f"Benchmarking {bench_name}")
    benchmark(config, results_folder, bench_name)
    print(f"Benchmark results wrote in {results_folder}")
